"use strict";

const anyMap = new WeakMap(), eventsMap = new WeakMap(), producersMap = new WeakMap(), anyProducer = Symbol("anyProducer"), resolvedPromise = Promise.resolve(), listenerAdded = Symbol("listenerAdded"), listenerRemoved = Symbol("listenerRemoved"), metaEventsAllowed = Symbol("metaEventsAllowed");

let isGlobalDebugEnabled = !1;

function assertEventName(e, t) {
if ("string" != typeof e && "symbol" != typeof e && "number" != typeof e) throw new TypeError("`eventName` must be a string, symbol, or number");
if (isMetaEvent(e) && t !== metaEventsAllowed) throw new TypeError("`eventName` cannot be meta event `listenerAdded` or `listenerRemoved`");
}

function assertListener(e) {
if ("function" != typeof e) throw new TypeError("listener must be a function");
}

function getListeners(e, t) {
const r = eventsMap.get(e);
r.has(t) || r.set(t, new Set());
return r.get(t);
}

function getEventProducers(e, t) {
const r = "string" == typeof t || "symbol" == typeof t || "number" == typeof t ? t : anyProducer, s = producersMap.get(e);
s.has(r) || s.set(r, new Set());
return s.get(r);
}

function enqueueProducers(e, t, r) {
const s = producersMap.get(e);
if (s.has(t)) for (const e of s.get(t)) e.enqueue(r);
if (s.has(anyProducer)) {
const e = Promise.all([ t, r ]);
for (const t of s.get(anyProducer)) t.enqueue(e);
}
}

function iterator(e, t) {
t = Array.isArray(t) ? t : [ t ];
let r = !1, s = () => {}, n = [];
const o = {
enqueue(e) {
n.push(e);
s();
},
finish() {
r = !0;
s();
}
};
for (const r of t) getEventProducers(e, r).add(o);
return {
async next() {
if (!n) return {
done: !0
};
if (0 === n.length) {
if (r) {
n = void 0;
return this.next();
}
await new Promise(e => {
s = e;
});
return this.next();
}
return {
done: !1,
value: await n.shift()
};
},
async return(r) {
n = void 0;
for (const r of t) getEventProducers(e, r).delete(o);
s();
return arguments.length > 0 ? {
done: !0,
value: await r
} : {
done: !0
};
},
[Symbol.asyncIterator]() {
return this;
}
};
}

function defaultMethodNamesOrAssert(e) {
if (void 0 === e) return allEmitteryMethods;
if (!Array.isArray(e)) throw new TypeError("`methodNames` must be an array of strings");
for (const t of e) if (!allEmitteryMethods.includes(t)) {
if ("string" != typeof t) throw new TypeError("`methodNames` element must be a string");
throw new Error(`${t} is not Emittery method`);
}
return e;
}

const isMetaEvent = e => e === listenerAdded || e === listenerRemoved;

class Emittery {
static mixin(e, t) {
t = defaultMethodNamesOrAssert(t);
return r => {
if ("function" != typeof r) throw new TypeError("`target` must be function");
for (const e of t) if (void 0 !== r.prototype[e]) throw new Error(`The property \`${e}\` already exists on \`target\``);
Object.defineProperty(r.prototype, e, {
enumerable: !1,
get: function() {
Object.defineProperty(this, e, {
enumerable: !1,
value: new Emittery()
});
return this[e];
}
});
const s = t => (function(...r) {
return this[e][t](...r);
});
for (const e of t) Object.defineProperty(r.prototype, e, {
enumerable: !1,
value: s(e)
});
return r;
};
}
static get isDebugEnabled() {
if ("object" != typeof process) return isGlobalDebugEnabled;
const {env: e} = process || {
env: {}
};
return "emittery" === e.DEBUG || "*" === e.DEBUG || isGlobalDebugEnabled;
}
static set isDebugEnabled(e) {
isGlobalDebugEnabled = e;
}
constructor(e = {}) {
anyMap.set(this, new Set());
eventsMap.set(this, new Map());
producersMap.set(this, new Map());
this.debug = e.debug || {};
void 0 === this.debug.enabled && (this.debug.enabled = !1);
this.debug.logger || (this.debug.logger = (e, t, r, s) => {
try {
s = JSON.stringify(s);
} catch {
s = `Object with the following keys failed to stringify: ${Object.keys(s).join(",")}`;
}
"symbol" != typeof r && "number" != typeof r || (r = r.toString());
const n = new Date(), o = `${n.getHours()}:${n.getMinutes()}:${n.getSeconds()}.${n.getMilliseconds()}`;
console.log(`[${o}][emittery:${e}][${t}] Event Name: ${r}\n\tdata: ${s}`);
});
}
logIfDebugEnabled(e, t, r) {
(Emittery.isDebugEnabled || this.debug.enabled) && this.debug.logger(e, this.debug.name, t, r);
}
on(e, t) {
assertListener(t);
e = Array.isArray(e) ? e : [ e ];
for (const r of e) {
assertEventName(r, metaEventsAllowed);
getListeners(this, r).add(t);
this.logIfDebugEnabled("subscribe", r, void 0);
isMetaEvent(r) || this.emit(listenerAdded, {
eventName: r,
listener: t
}, metaEventsAllowed);
}
return this.off.bind(this, e, t);
}
off(e, t) {
assertListener(t);
e = Array.isArray(e) ? e : [ e ];
for (const r of e) {
assertEventName(r, metaEventsAllowed);
getListeners(this, r).delete(t);
this.logIfDebugEnabled("unsubscribe", r, void 0);
isMetaEvent(r) || this.emit(listenerRemoved, {
eventName: r,
listener: t
}, metaEventsAllowed);
}
}
once(e) {
let t;
const r = new Promise(r => {
t = this.on(e, e => {
t();
r(e);
});
});
r.off = t;
return r;
}
events(e) {
e = Array.isArray(e) ? e : [ e ];
for (const t of e) assertEventName(t, metaEventsAllowed);
return iterator(this, e);
}
async emit(e, t, r) {
assertEventName(e, r);
this.logIfDebugEnabled("emit", e, t);
enqueueProducers(this, e, t);
const s = getListeners(this, e), n = anyMap.get(this), o = [ ...s ], i = isMetaEvent(e) ? [] : [ ...n ];
await resolvedPromise;
await Promise.all([ ...o.map(async e => {
if (s.has(e)) return e(t);
}), ...i.map(async r => {
if (n.has(r)) return r(e, t);
}) ]);
}
async emitSerial(e, t, r) {
assertEventName(e, r);
this.logIfDebugEnabled("emitSerial", e, t);
const s = getListeners(this, e), n = anyMap.get(this), o = [ ...s ], i = [ ...n ];
await resolvedPromise;
for (const e of o) s.has(e) && await e(t);
for (const r of i) n.has(r) && await r(e, t);
}
onAny(e) {
assertListener(e);
this.logIfDebugEnabled("subscribeAny", void 0, void 0);
anyMap.get(this).add(e);
this.emit(listenerAdded, {
listener: e
}, metaEventsAllowed);
return this.offAny.bind(this, e);
}
anyEvent() {
return iterator(this);
}
offAny(e) {
assertListener(e);
this.logIfDebugEnabled("unsubscribeAny", void 0, void 0);
this.emit(listenerRemoved, {
listener: e
}, metaEventsAllowed);
anyMap.get(this).delete(e);
}
clearListeners(e) {
e = Array.isArray(e) ? e : [ e ];
for (const t of e) {
this.logIfDebugEnabled("clear", t, void 0);
if ("string" == typeof t || "symbol" == typeof t || "number" == typeof t) {
getListeners(this, t).clear();
const e = getEventProducers(this, t);
for (const t of e) t.finish();
e.clear();
} else {
anyMap.get(this).clear();
for (const e of eventsMap.get(this).values()) e.clear();
for (const e of producersMap.get(this).values()) {
for (const t of e) t.finish();
e.clear();
}
}
}
}
listenerCount(e) {
e = Array.isArray(e) ? e : [ e ];
let t = 0;
for (const r of e) if ("string" != typeof r) {
"undefined" != typeof r && assertEventName(r, metaEventsAllowed);
t += anyMap.get(this).size;
for (const e of eventsMap.get(this).values()) t += e.size;
for (const e of producersMap.get(this).values()) t += e.size;
} else t += anyMap.get(this).size + getListeners(this, r).size + getEventProducers(this, r).size + getEventProducers(this).size;
return t;
}
bindMethods(e, t) {
if ("object" != typeof e || null === e) throw new TypeError("`target` must be an object");
t = defaultMethodNamesOrAssert(t);
for (const r of t) {
if (void 0 !== e[r]) throw new Error(`The property \`${r}\` already exists on \`target\``);
Object.defineProperty(e, r, {
enumerable: !1,
value: this[r].bind(this)
});
}
}
}

const allEmitteryMethods = Object.getOwnPropertyNames(Emittery.prototype).filter(e => "constructor" !== e);

Object.defineProperty(Emittery, "listenerAdded", {
value: listenerAdded,
writable: !1,
enumerable: !0,
configurable: !1
});

Object.defineProperty(Emittery, "listenerRemoved", {
value: listenerRemoved,
writable: !1,
enumerable: !0,
configurable: !1
});

window.Emittery = Emittery;